"""
``self-paced ensemble`` is a python-based ensemble learning framework for 
dealing with binary class-imbalanced classification problems in machine learning.
"""
# License: MIT

__version__ = '0.1.7'